@@ -9,7 +9,22 @@ include("testsetup.jl")
   println("Aqua.test_all")
   t0 = time_ns()
   deps_compat = VERSION <= v"1.8" || isempty(VERSION.prerelease)
-  Aqua.test_all(VectorizationBase; deps_compat = deps_compat)
+
+  # TODO - Will need a code refactor to properly address these type piracies.
+  #        Either:
+  #        1. Create type wrappers in VectorizationBase 
+  #        2. Implement overloading upstream
+  #        3. Use package extensions (still buggy in current Julia LTS v1.10.10)
+
+  pirated_types = [
+            VectorizationBase.FastRange,
+            VectorizationBase.AbstractStridedPointer,
+            VectorizationBase.StridedBitPointer,
+            VectorizationBase.StaticInt,
+            VectorizationBase.AbstractSIMD,
+            VectorizationBase.Bit,
+        ]
+  Aqua.test_all(VectorizationBase; deps_compat = deps_compat, piracies=(treat_as_own = pirated_types,))
   println("Aqua took $((time_ns() - t0)*1e-9) seconds")
   # @test isempty(detect_unbound_args(VectorizationBase))
   # @test isempty(detect_ambiguities(VectorizationBase))
